import numpy as np
from tqdm import tqdm
from datetime import datetime, timedelta
from joblib import Parallel, delayed
from scipy.signal import detrend


def par_corr_npy(data, threshold=0.5):
    rows, cols = data.shape
    data = data[~np.isnan(data).any(axis=1)]
    if data.shape[0] < rows*threshold:
        return np.full(cols-1, np.nan)
    
    data = detrend(data, axis=0)
    # data = zscore(data, axis=0)
    corr_matrix = np.corrcoef(data, rowvar=False)
    inv_matrix = np.linalg.inv(corr_matrix)
    r = np.array([-inv_matrix[0, col] / np.sqrt(inv_matrix[0, 0] * inv_matrix[col, col]) for col in range(1, cols)])
    return r


def partial_corr(x, y, threshold=0.5, n_jobs=-1):
    data = np.concatenate([y[..., np.newaxis], x], axis=-1) # (660000, 21, 4)
    result = np.array(Parallel(n_jobs=n_jobs)(delayed(par_corr_npy)(data[i], threshold) for i in range(data.shape[0]))) # (660000, 3)
    return result



def get_opl(y:np.ndarray, x:np.ndarray, **kwargs):
    """
    Calculate the optimal pre-season length for specified variables.

    Parameters:
    - y: np.ndarray, shape (rows, cols, years, ) 
      Target variable data over the years. such as phenological data at an annual scale.

    - x: np.ndarray, shape (rows, cols, months, xvars)
      Explanatory variables data at a monthly scale, covering one more year than y data (months = 12 * (years + 1)).

    Keyword Arguments (kwargs):
    - n_jobs: int, default --> mp.cpu_count()
      number of jobs to run in parallel. 
    
    - max_month: int, default 6
      Maximum pre-season length calculated forward from the current month.


    Returns: 
    - for multiple data (eg. x.ndim == 4)
      result: np.ndarray, shape (rows, cols, xvars) the value of opl for each variable.
    """
    n_jobs = kwargs.get('n_jobs', -1)
    max_month = kwargs.get('max_month', 6)
    threshold = kwargs.get('threshold', 0.5)
    
    # 形状适配
    y = y.squeeze()
    x = x.squeeze()
    if x.ndim == 2:
        x = x[np.newaxis, np.newaxis, :, :]  # (1, 1, months, xvars)
        y = y[np.newaxis, np.newaxis, :]  # (1, 1, years)

    elif x.ndim == 3:
        x = x[np.newaxis, :, :, :]  # (1, cols, months, xvars)
        y = y[np.newaxis, :, :]  # (1, cols, years)
        
    
    rows, cols, months, xvars = x.shape
    _, _, years = y.shape

    mask_y = np.sum(~np.isnan(y), axis=2) >= years*threshold # (600, 3600)
    mask_x = np.sum(~np.isnan(x), axis=2) >= months*threshold # (600, 3600, 3)
    mask_x = np.sum(mask_x, axis=2) == xvars # (600, 3600)
    mask = mask_y & mask_x # (600, 3600)

    y = y[mask] # (660000, 21)
    x = x[mask] # (660000, 264, 3)

    # 获取多年平均物候期所在月份
    y_mean = np.nanmean(y, axis=-1) # (660000, )
    y_mean = np.array([(datetime(2023, 1, 1) + timedelta(days=int(round(day)))).month for day in y_mean])
    y_mean = np.expand_dims(y_mean, axis=-1).astype(int) # (660000, 1) 全是5月


    indices1 = np.arange(years) * 12 + y_mean - 1  # (660000, 21)
    indices0 = np.arange(x.shape[0])[:, np.newaxis]

    x_list = [x[:, 12-i:-i if i!= 0 else None, :] for i in range(max_month)]
    x_list = [np.mean(x_list[:i+1], axis=0)[indices0, indices1, :] for i in range(max_month)]

    
    result = [partial_corr(x, y, threshold, n_jobs) for x in tqdm(x_list)]
    result = np.abs(np.array(result)) # (6, 660000, 3)
    result = np.argmax(result, axis=0) # (660000, 3)
    

    result_mask = np.full((rows, cols, xvars), np.nan)
    result_mask[mask] = result
    result_mask = result_mask.squeeze()
    
    return result_mask



# -------------------------------------------------------------------------- #

    
def single_mean(x, opl, pheno):
    """
    x: shape (months, )
    opl: number
    pheno: number
    """
    k_indices = np.arange(12, len(x), 12) # (months//12-1, )
    start_indices = (k_indices + pheno - opl - 1).astype(int) # (months//12-1, )
    end_indices = (k_indices + pheno).astype(int) # (months//12-1, )
    mask = (np.arange(len(x)) >= start_indices[:, np.newaxis]) & (np.arange(len(x)) < end_indices[:, np.newaxis])
    x_masked = np.where(mask, x, np.nan) 
    resarr = np.nanmean(x_masked, axis=1) # shape (months//12-1, )
    
    return resarr


def seasonal_mean(x_arr: np.ndarray, opl_arr: np.ndarray, pheno_arr: np.ndarray):
    """
    Calculate pre-seasonal mean from input arrays.

    Parameters:
    - x_arr (np.ndarray): Climate variable array (months, rows, cols), where months is a multiple of 12.
    - opl_arr (np.ndarray): The optimal preseason length (rows, cols).
    - pheno_arr (np.ndarray): Phenological dates (rows, cols), typically multi-year average in day of year (1-365) or month (1-12).

    Returns:
    - result (np.ndarray): Pre-seasonal mean (months//12-1, rows, cols).

    """

    if np.nanmax(pheno_arr) > 12: # if the value of pheno_arr is the "day of year" (1-365), then convert it to the month
        pheno_arr[~np.isnan(pheno_arr)] = np.array([(datetime(2023, 1, 1) + timedelta(days=eos - 1)).month for eos in pheno_arr[~np.isnan(pheno_arr)]])

    months, rows, cols = x_arr.shape
 
    nn_indices = np.where(~np.isnan(opl_arr) & ~np.isnan(pheno_arr)) # nn_indices, means the indices of non-NaN elements
    nn_opl = opl_arr[nn_indices] # shape (806542,)
    nn_pheno = pheno_arr[nn_indices] # shape (806542,)
    nn_x = x_arr[:, nn_indices[0], nn_indices[1]] # shape (504, 806542)

    # slice the data
    args_list = [(nn_x[:, i], nn_opl[i], nn_pheno[i]) for i in range(nn_opl.shape[0])]
    res = np.array(Parallel(n_jobs=-1)(delayed(single_mean)(x, opl, pheno) for x, opl, pheno in tqdm(args_list))) # shape (806542, 41)
    result = np.full((months//12-1, rows, cols), np.nan)
    result[:, nn_indices[0], nn_indices[1]] = res.swapaxes(0, 1)

    return result
 
    

    
            



import xarray as xr


def swap_longtitude(ds:xr.Dataset, **kwargs):
    '''
    Converts longitudes based on the input dataset's longitude range:
    - If the input dataset's longitude range is -180 to 180 degrees, it converts to 0-360 degrees.
    - If the input dataset's longitude range is 0-360 degrees, it converts to -180 to 180 degrees.

    Parameters:
    - ds (xarray.Dataset): Input dataset.
    - lon_dim (str, optional): The dimension representing longitudes. Default is 'longitude'.

    Returns:
    xarray.Dataset: Dataset with longitudes converted based on the specified conditions and sorted by latitude and longitude.
    '''
    # 自动获取lon_dim
    lon_dim = kwargs.get('lon_dim', 'lon')
    swap_type = kwargs.get('swap_type', '180')
    dims = list(ds.dims)
    if not lon_dim in dims:
        for item in ['longitude', 'x', 'Lon', 'Longitude']:
            if item in dims:
                lon_dim = item
                break
    if not lon_dim in dims:
        lon_dim = max(dims, key=dims.get)

    lon_arr = ds[lon_dim].values
    if swap_type == '180' and max(ds[lon_dim].values) > 300:
        lon_arr[lon_arr > 180] = lon_arr[lon_arr > 180] - 360
    elif swap_type == '360' and min(ds[lon_dim].values) < 0:
        lon_arr[lon_arr < 0] = lon_arr[lon_arr < 0] + 360

    ds[lon_dim] = lon_arr

    return ds.sortby(lon_dim)



import os
import cdsapi
import calendar
import numpy as np
import pandas as pd
import xarray as xr
from shapely.geometry import Point
import geopandas as gpd
from pyproj import CRS
from subprocess import call


def idm_era5(idm_engine, variable, dst_dir, **kwargs):
    retrieve_type = kwargs.get('retrieve_type', 'reanalysis-era5-land-monthly-means')
    product_type = kwargs.get('product_type','monthly_averaged_reanalysis')
    year = kwargs.get('year', [str(year) for year in range(1980, 2024)])
    month = kwargs.get('month', [str(mon).zfill(2) for mon in range(1, 13)])
    time = kwargs.get('time', '00:00')
    format = kwargs.get('format', 'netcdf.zip')
    file_name = kwargs.get('file_name', f'{variable}.zip')
    
    c = cdsapi.Client()
    dic = {
        'product_type': product_type,
        'variable': variable,
        'year': year,
        'month': month,
        'time': time,
        'format': format
    }

    file_path = os.path.join(dst_dir, file_name)
    
    if not os.path.exists(file_path):
        r = c.retrieve(retrieve_type, dic)
        task_url = r.location
        print('正在下载%s'%task_url)
        call([idm_engine, '/d', task_url, '/p', dst_dir, '/f', file_name, '/a'])
        call([idm_engine, '/s'])
        
    else:
        print(file_path, "存在同名文件")
        
        
def idm_era5_daily(idm_engine, variable, dst_dir, **kwargs):
    retrieve_type = kwargs.get('retrieve_type','reanalysis-era5-land')
    year = kwargs.get('year', '2023')
    month = kwargs.get('month', '01')
    day = kwargs.get('day', [str(day).zfill(2) for day in range(1, calendar.monthrange(year, month)[1] + 1)])
    time = kwargs.get('time', [f"{str(hour).zfill(2)}:00" for hour in range(24)])
    format = kwargs.get('format', 'netcdf.zip')
    file_name = kwargs.get('file_name', f'{variable}.zip')
    
    c = cdsapi.Client()
    dic = {
        'variable': variable,
        'year': year,
        'month': month,
        'day': day,
        'time': time,
        'format': format
    }

    file_path = os.path.join(dst_dir, file_name)
    
    if not os.path.exists(file_path):
        r = c.retrieve(retrieve_type, dic)
        task_url = r.location
        print('正在下载%s'%task_url)
        call([idm_engine, '/d', task_url, '/p', dst_dir, '/f', file_name, '/a'])
        call([idm_engine, '/s'])
        
    else:
        print(file_path, "存在同名文件")
        

def truncate_and_unique_column_names(data, max_length=10):
    """
    截断DataFrame的列名为10个字符以内并确保唯一性。
    
    Parameters:
    data (pd.DataFrame): 输入的DataFrame。

    Returns:
    pd.DataFrame: 具有截断且唯一列名的DataFrame。
    """
    truncated_columns = []
    for col in data.columns:
        truncated_col = col[:max_length]  # 截断为10个字符以内
        suffix = 1
        while truncated_col in truncated_columns:
            # 处理重复列名，添加数字后缀
            truncated_col = col[:max_length - len(str(suffix))] + str(suffix)
            suffix += 1
        truncated_columns.append(truncated_col)

    data.columns = truncated_columns
    return data

def df2shp(df, file_path, lon_name='longitude', lat_name='latitude'):
    longs = df[lon_name].values
    if np.nanmax(np.abs(longs)) > 190:
        df1 = df.copy()[df[lon_name] > 180]
        df1[lon_name] = df1[lon_name] - 360
        df2 = df.copy()[df[lon_name] <= 180]
        df = pd.concat([df1, df2], axis=0)
        
    df = truncate_and_unique_column_names(df)
    # 将经纬度列转换为Point类型的几何列
    geometry = [Point(xy) for xy in zip(df[lon_name], df[lat_name])]
    # 创建GeoDataFrame对象并设置几何列
    gdf = gpd.GeoDataFrame(df, geometry=geometry)
    # 将坐标系设置为WGS84
    gdf.crs = CRS.from_epsg(4326)
    # 将数据保存为.shp文件
    gdf.to_file(file_path)
    
    
def dfsummary(df, r_name='r', p_name ='p-val', oprint=True):
    df = df[[r_name, p_name]].dropna()
    
    c1 = len(df)
    p1 = len(df[df[r_name]>0])/c1
    n1 = len(df[df[r_name]<0])/c1

    df = df[df[p_name]<0.05]
    c2 = len(df)
    if c2 > 0:
        p2 = len(df[df[r_name]>0])/c1
        n2 = len(df[df[r_name]<0])/c1
        
        df = df[df[p_name]<0.01]
        c3 = len(df)
        if c3>0:
            p3 = len(df[df[r_name]>0])/c1
            n3 = len(df[df[r_name]<0])/c1
        else:
            c3 = 0
            p3 = 0
            n3 = 0
    else:
        c3 = p2 = n2 = p3 = n3 = 0


    pn1 = round(p1/n1, 2) if n1 != 0 else '-'
    pn2 = round(p2/n2, 2) if n2 != 0 else '-'
    pn3 = round(p3/n3, 2) if n3 != 0 else '-'

    # 定义 ANSI 转义码
    ITALIC_START = '\033[3m'
    ITALIC_END = '\033[0m'

    c_len = max(7, len(str(int(c1)))+1)
    if oprint is True:
        # 打印文本，将 P<0.05 设置为斜体
        print(f"{ITALIC_START}{'Summary':>9} {'Positive':>9} {'Negative':>9} {'P:N':>5} {'Count':>{c_len}} \
            \n{'Total':>8}: {p1*100:>7.2f}%  {n1*100:>7.2f}% {pn1:>6} {c1:>{c_len}} \
            \n{'*P<0.05':>8}: {p2*100:>7.2f}%  {n2*100:>7.2f}% {pn2:>6} {c2:>{c_len}} \
            \n{'**P<0.01':>8}: {p3*100:>7.2f}%  {n3*100:>7.2f}% {pn3:>6} {c3:>{c_len}}{ITALIC_END}"
        )
    return p1, p2, p3, n1, n2, n3







def remove_outliers(data, threshold=3):
    """
    Function to remove outliers from a given data array.

    Parameters:
    - data: One-dimensional or multi-dimensional data array, can be a pandas Series, NumPy array, or xarray.DataArray.
    - threshold: Threshold for outliers in terms of the data standard deviation. Default is 3.

    Returns:
    - Processed data array with outliers beyond the threshold replaced with NaN.

    Note:
    - For NumPy arrays, the function directly modifies the input array, replacing outliers with NaN.
    - For xarray.DataArray, a new array is generated where outliers are replaced with NaN, and the original array remains unaffected.
    - If the input is a pandas Series, a new Series is generated with outliers replaced with NaN, and the original Series remains unaffected.

    """
    
    # Calculate the standard deviation, mean, and center the data
    data_std = np.nanstd(data)
    data_mean = np.nanmean(data)
    data_centered = np.abs(data - data_mean)
    
    # Remove outliers based on the threshold and data type
    if isinstance(data, xr.DataArray):
        data = xr.where(data_centered > threshold*data_std, np.nan, data)
    else:
        data[data_centered > threshold*data_std] = np.nan

    return data


import numpy as np
from scipy.stats import mode
from .api_resample import resample_2D_mean, resample_3D_mean
from joblib import Parallel, delayed
import math
import numpy as np
from tqdm import tqdm


def aggregate(x, exc_shape, ignore_nodata=None):
    """
    x: 2D array
    exc_shape: tuple of (rows, cols)
    return: 2D array
    """
    cur_rows, cur_cols = x.shape
    exc_rows, exc_cols = exc_shape
    rw = cur_rows / exc_rows
    cw = cur_cols / exc_cols
    
    x = np.pad(x, ((0, int(rw+1)), (0, int(cw+1))), 'constant', constant_values=np.nan)
    x[np.isinf(x)] = np.nan
    
    if ignore_nodata is not None:
        x[x==ignore_nodata] = np.nan
    
    x_min = np.nanmin(x)-1
    x = x - x_min # x >= 1

    x[np.isnan(x)] = 0 # 0 is the missing value
    x = x.astype(np.float64)
    y = np.full(exc_shape, np.nan).astype(np.float64)
    y = resample_2D_mean(x, y, rw, cw, exc_rows, exc_cols, 0) # 0 is the missing value
    y = np.asarray(y) + x_min
    
    return y


def aggregate3D(x, exc_shape, ignore_nodata=None):
    """
    x: 2D array
    exc_shape: tuple of (rows, cols)
    return: 2D array
    """
    cur_times, cur_rows, cur_cols = x.shape
    exc_times, exc_rows, exc_cols = exc_shape
    rw = cur_rows / exc_rows
    cw = cur_cols / exc_cols
    
    x = np.pad(x, (0, (0, int(rw+1)), (0, int(cw+1))), 'constant', constant_values=np.nan)
    x[np.isinf(x)] = np.nan
    
    if ignore_nodata is not None:
        x[x==ignore_nodata] = np.nan
    
    x_min = np.nanmin(x)-1
    x = x - x_min # x >= 1

    x[np.isnan(x)] = 0 # 0 is the missing value
    x = x.astype(np.float64)
    y = np.full(exc_shape, np.nan).astype(np.float64)
    y = resample_3D_mean(x, y, rw, cw, exc_times, exc_rows, exc_cols, 0) # 0 is the missing value
    y = np.asarray(y) + x_min
    
    return y

def aggregate3D_v2(data, exc_shape, n_jobs=-1):
    res = Parallel(n_jobs)(delayed(aggregate)(data_i, exc_shape) for data_i in data)
    res = np.array(res)
    return res


def nanmode(data, axis=-1):
    mask = np.sum(np.isnan(data), axis=axis)==0
    data_np = data[mask].astype(np.int32)
    data_sy = data[~mask] # (1, 36)

    # 求第三个维度上的众数
    if data_np.size > 0:
        res_np = np.apply_along_axis(lambda x: np.argmax(np.bincount(x)), axis=axis, arr=data_np)
    else:
        res_np = None
        
    if data_sy.size > 0:
        res_sy = mode(data_sy, axis=axis, nan_policy='omit').mode
    else:
        res_sy = None

    result = np.full(mask.shape, np.nan)
    result[mask] = res_np
    result[~mask] = res_sy
    
    return result


def slide_win_npy(data, step=10, winsize=10):
    l_ws = winsize//2 # 中心点左和上侧边长
    r_ws = winsize - l_ws # 中心点右和下侧边长
    height, width, bands = data.shape
    
    x_range = np.arange(l_ws, height+l_ws, step).astype(int)
    y_range = np.arange(l_ws, width+l_ws, step).astype(int)
    data = np.concatenate((data, data[:, :winsize, :]), axis=1) # (2400, 14410, 13)
    data = np.pad(data, ((0, winsize), (0, 0), (0, 0)), mode='constant', constant_values=np.nan) # (2410, 14410, 13)
    
    res_list = []
    for i in x_range:
        i_start = i - l_ws
        i_end = i + r_ws
        for j in y_range:
            j_start = j - l_ws
            j_end = j + r_ws
            res_list.append(data[i_start:i_end, j_start:j_end])

    res = np.array(res_list).reshape(len(x_range), len(y_range), winsize, winsize, bands)
    return res



def reshape_2d_4d(data, exc_rows, exc_cols):
    rows, cols = data.shape
    scale_r = int(rows / exc_rows)
    scale_c = int(cols / exc_cols)
    
    data = data.reshape(exc_rows, scale_r, cols).transpose(2, 0, 1) 
    data = data.reshape(exc_cols, scale_c, exc_rows, scale_r).transpose(2, 0, 3, 1) # (exc_rows, exc_cols, scale_r, scale_c)
    # data = data.reshape(exc_rows, exc_cols, -1)
    return data



class Aggregate:

    def __init__(self, data: np.ndarray, exc_shape:tuple):
        '''
        Initialize Aggregator object.

        Parameters:
        data (numpy.ndarray): The input data to be aggregated.
        exc_shape (tuple): The expected shape of the output data.
        '''
        self.data = data
        self.exc_shape = exc_shape
        self.exc_rows, self.exc_cols = self.exc_shape
        self.cur_rows, self.cur_cols = self.data.shape
        self.r_res, self.c_res = self.cur_rows/self.exc_rows, self.cur_cols/self.exc_cols

        pass
        
    def mode(self): return self.fit('mode')
    def mean(self): return self.fit('mean')

    
    def fit(self, operation: str = 'mode'):
        '''
        Aggregate the data using mode operation.

        Returns:
        numpy.ndarray: The aggregated data.
        '''
       
        if self.r_res.is_integer() and self.c_res.is_integer():
            self.data = self.data.reshape(self.exc_rows, int(self.r_res), self.exc_cols, int(self.c_res))
            self.data = self.data.transpose(0, 2, 1, 3).reshape(self.exc_rows, self.exc_cols, -1)
            if operation == 'mode':
                return self._mode_exa_div()
            elif operation == 'mean':
                return np.nanmean(self.data, axis=-1)
            else:
                raise ValueError("Invalid operation")
            
        else:
            return self._mode_not_div(operation)




    def _get_boundary(self, r, ratio):
        r_sta = r*ratio
        r_end = (r+1)*ratio

        rmin = int(r_sta)
        rmax = math.ceil(r_end)

        offset_1 = math.ceil(r_sta) - r_sta # 上/左 偏移量
        offset_2 = r_end - int(r_end) # 下/右 偏移量
        return rmin, rmax, offset_1, offset_2
    


    def _get_area(self, win_shape, offsets):

        offset_1, offset_2, offset_3, offset_4 = offsets # 上下左右
        areas = np.full(win_shape, 1, dtype=float)
        if offset_1 > 0: areas[ 0, :] = offset_1 
        if offset_2 > 0: areas[-1, :] = offset_2
        if offset_3 > 0: areas[:,  0] = offset_3 * areas[:,  0]
        if offset_4 > 0: areas[:, -1] = offset_4 * areas[:, -1]

        return areas



    def _mode_not_div(self, operation):
        result = np.full(self.exc_shape, np.nan)
        for r in tqdm(range(self.exc_rows)):
            rmin, rmax, offset_1, offset_2 = self._get_boundary(r, self.r_res)
            
            for c in range(self.exc_cols):
                cmin, cmax, offset_3, offset_4 = self._get_boundary(c, self.c_res)
                data_win = self.data[rmin:rmax, cmin:cmax]
                mask = ~np.isnan(data_win)
                if not mask.any():
                    continue

                area_win = self._get_area(data_win.shape, (offset_1, offset_2, offset_3, offset_4))
                data_ij = data_win[mask].astype(np.int64)
                area_ij = area_win[mask]

                if operation == 'mode':
                    counts = np.bincount(data_ij, weights=area_ij)*1e9
                    result[r, c] = np.argmax(counts.astype(np.int64)) # areas整型或浮点型 对np.argmax(areas)的结果有影响。具体影响在于当多个val并列第一的时候, 该取哪一个? 据简单观测, 整型-->取最小的, 浮点型-->取最大
                elif operation == 'mean':
                    result[r, c] = np.sum(data_ij*area_ij)/np.sum(area_ij)
        
        return result


    def _mode_exa_div(self,):
        mask = np.any(~np.isnan(self.data), axis=-1)
        result = np.full(mask.shape, np.nan)
        result[mask] = np.apply_along_axis(lambda x: np.argmax(np.bincount(np.int64(x[~np.isnan(x)]))), axis=-1, arr=self.data[mask])
        return result
    





import os
import sys
import setuptools

# Get the compatible .pyd files for python version
package_dir = os.path.join(os.getcwd(), "phenology")
if not os.path.exists(package_dir):
    package_dir = os.getcwd()
    print("Package directory not found. Using current directory instead.")
    
pyd_list = [os.path.join(root, file) for root, dirs, files in os.walk(package_dir)
            for file in files if file.endswith(".pyd") and f"cp{sys.version_info.major}{sys.version_info.minor}" in file]

# Get the content of the README.md file
with open("README.md", "r", encoding = "utf-8") as fh:
    long_description = fh.read()

# Set up the package using setuptools
setuptools.setup(
    name = "phenology",
    version = "0.0.13",
    author = "Shen Pengju",
    author_email = "spjace@sina.com",
    description = "A small package for phenology analysis",
    long_description = long_description,
    long_description_content_type = "text/markdown",
    url = "https://github.com/spjace/phenology",
    packages = setuptools.find_packages(),
    install_requires=["cdsapi", "geopandas", "joblib", "numpy", "pandas", "pyproj", "scipy", "Shapely", "tqdm", "xarray"],
    classifiers = [
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7",
    include_package_data=True,
    package_data={"phenology": pyd_list},
)



import numpy as np
import pandas as pd
import multiprocessing as mp
from tqdm import tqdm
from .api_opl_v1 import opl_api


def opl_for_parallel(args):
    y, x, max_month, target_vars, index = args
    result = opl_api(y, x, max_month=max_month, target_vars=target_vars, index=index)
    return result
  

def get_opl_v1(y:np.ndarray, x:np.ndarray, **kwargs):
    """
    Calculate the optimal pre-season length for specified variables.

    Parameters:
    - y: np.ndarray, shape (rows, cols, years, 1) or (years, 1)
      Target variable data over the years. such as phenological data at an annual scale.

    - x: np.ndarray, shape (rows, cols, months, vars) or (months, vars)
      Explanatory variables data at a monthly scale, covering one more year than y data (months = 12 * (years + 1)).

    Keyword Arguments (kwargs):
    - n_jobs: int, default --> mp.cpu_count()
      number of jobs to run in parallel. 
    
    - max_month: int, default 6
      Maximum pre-season length calculated forward from the current month.

    - target_vars: list, default None
      variable names for calculating optimal pre-season length. Default --> x.columns.to_list().

    Returns: 
    - for multiple data (eg. x.ndim == 4)
      result: np.ndarray, shape (rows, cols, 6, len(target_vars))
      In the 3rd dimension of result, the 6 elements are ['n', 'r', 'ci95lower', 'ci95upper', 'p-val', 'opl'], where opl is the optimal pre-season length.
      
    - for single data (eg. x.ndim == 2)
      result: pd.DataFrame
      Columns: ['n', 'r', 'CI95%', 'p-val', 'opl', 'var', 'index'], where opl is the optimal preseason length.
    """
    

    n_jobs = kwargs.get('n_jobs', mp.cpu_count())
    max_month = kwargs.get('max_month', 6)
    target_vars = kwargs.get('target_vars', [f'x{i}' for i in range(x.shape[-1])])
    
    
    if x.ndim == 2:
      if y.ndim == 1:
        y = y.reshape((-1, 1))
      return opl_api(y, x, max_month=max_month, target_vars=target_vars)
    

    rows, cols, months, vars  = x.shape
    x = x.reshape((rows * cols, months, vars)) 
    y = y.reshape((rows * cols, -1, 1))

    args_list = [[y[index], x[index], max_month, target_vars, index] for index in range(rows * cols)]
    
    print('Start parallel computation, parallel processing cores: ', n_jobs)
    
    result = []
    with mp.Pool(processes=n_jobs) as pool:
        for res in tqdm(pool.imap_unordered(opl_for_parallel, args_list), ncols=100, total=len(args_list)):
            result.append(res)
    
    result = pd.concat(result, axis=0)
    
    # Split CI95% column
    result[['ci95lower', 'ci95upper']] = pd.DataFrame(result['CI95%'].tolist(), index=result.index)
    result.drop('CI95%', axis=1, inplace=True)
    
    # Identify missing rows
    missing_rows = pd.DataFrame(columns=result.columns)
    missing_rows['index'] = list(set(range(rows * cols)).difference(result['index']))

    var_list = result['var'].unique()

    result_list = [result] + [missing_rows.assign(var=i) for i in var_list]
    res_grouped = pd.concat(result_list).groupby('var')

    result = np.stack([group_data.sort_values('index')[['n', 'r', 'ci95lower',	'ci95upper', 'p-val', 'opl']].values
                       .reshape(rows, cols, -1)
                       .astype(np.float32)
                       for group_id, group_data in res_grouped
                       ], axis=-1
                      )
    return result

